/*
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <windows.h>
#include <dbghelp.h>

// =======================================================================================================================================================================================================================================================
// WINDOWS NT - KERNEL INTERACTION
// =======================================================================================================================================================================================================================================================

typedef enum _PROCESSINFOCLASS
{
  ProcessBasicInformation,  //  0
}
PROCESSINFOCLASS;

typedef struct _PROCESS_BASIC_INFORMATION // Information Class 0
{
  DWORD ExitStatus;
  LPVOID PebBaseAddress;
  ULONG_PTR AffinityMask;
  DWORD BasePriority;
  ULONG_PTR UniqueProcessId;
  ULONG_PTR InheritedFromUniqueProcessId;
}
PROCESS_BASIC_INFORMATION, *PPROCESS_BASIC_INFORMATION;

typedef DWORD (WINAPI *NTQUERYINFORMATIONPROCESS)(
  HANDLE ProcessHandle,
  PROCESSINFOCLASS ProcessInformationClass,
  PVOID ProcessInformation,
  ULONG ProcessInformationLength,
  PULONG ReturnLength
);

static DWORD WINAPI ThreadAlive(LPVOID Parameter)
{
  PROCESS_BASIC_INFORMATION Info;

  NTQUERYINFORMATIONPROCESS NtQueryInformationProcess = (NTQUERYINFORMATIONPROCESS) GetProcAddress(GetModuleHandle("ntdll.dll"), "NtQueryInformationProcess");
  NtQueryInformationProcess(GetCurrentProcess(), ProcessBasicInformation, (LPVOID) &Info, sizeof(PROCESS_BASIC_INFORMATION), NULL);

  WaitForSingleObject(OpenProcess(SYNCHRONIZE, FALSE, Info.InheritedFromUniqueProcessId), INFINITE);

  ExitProcess(9001);
  return 9001;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - DEBUGGER - CRASH
// =======================================================================================================================================================================================================================================================

typedef BOOL (WINAPI *MINIDUMPWRITEDUMP)(
        HANDLE hProcess,
        DWORD ProcessId,
        HANDLE hFile,
        MINIDUMP_TYPE DumpType,
        CONST PMINIDUMP_EXCEPTION_INFORMATION ExceptionParam,
        CONST PMINIDUMP_USER_STREAM_INFORMATION UserEncoderParam,
        CONST PMINIDUMP_CALLBACK_INFORMATION CallbackParam);

static MINIDUMPWRITEDUMP MiniDumpWriteDumpEx;

static BOOL CALLBACK MiniDumpCallback(PVOID Callback, CONST PMINIDUMP_CALLBACK_INPUT Input, PMINIDUMP_CALLBACK_OUTPUT Output)
{
  if (Input->CallbackType == IncludeModuleCallback)
  {
    BOOL Mine = (HINSTANCE) Input->IncludeModule.BaseOfImage == GetModuleHandle(NULL);

    if (!Mine)
      Output->ModuleWriteFlags &= ~(ModuleWriteCodeSegs | ModuleWriteDataSeg | ModuleWriteTlsData);
  }

  return TRUE;  (void) Callback;
}

static long __stdcall DebugExceptionFilter(LPEXCEPTION_POINTERS ExceptionPointers)
{
  static DWORD Counter;

  if (Counter++ > 0)
    return EXCEPTION_EXECUTE_HANDLER;

  char TempPath[MAX_PATH], TempName[MAX_PATH];

  GetTempPath(_countof(TempPath), TempPath);
  strcpy(TempName, TempPath);
  strcat(TempName, "\\drone.libav.dmp");

  MINIDUMP_EXCEPTION_INFORMATION Param;
  Param.ThreadId = GetCurrentThreadId();
  Param.ExceptionPointers = ExceptionPointers;
  Param.ClientPointers = TRUE;

  MINIDUMP_CALLBACK_INFORMATION Callback;
  Callback.CallbackRoutine = MiniDumpCallback;
  Callback.CallbackParam = NULL;

  DWORD DumpType =
    MiniDumpWithDataSegs |
    // MiniDumpWithFullMemory
    MiniDumpWithHandleData |
    // MiniDumpFilterMemory
    MiniDumpScanMemory |
    MiniDumpWithUnloadedModules |
    // MiniDumpWithIndirectlyReferencedMemory
    // MiniDumpFilterModulePaths
    MiniDumpWithProcessThreadData |
    // MiniDumpWithPrivateReadWriteMemory
    // MiniDumpWithoutOptionalData
    MiniDumpWithFullMemoryInfo |
    MiniDumpWithThreadInfo |
    MiniDumpWithCodeSegs;
  // MiniDumpWithoutAuxiliaryState
  // MiniDumpWithFullAuxiliaryState
  // MiniDumpWithPrivateWriteCopyMemory
  //- MiniDumpIgnoreInaccessibleMemory
  //- MiniDumpWithTokenInformation;

#if _WIN32_WINNT >= 0x0601

  DWORD MajorVersion = LOBYTE(LOWORD(GetVersion()));

  if (MajorVersion < 6) DumpType &= ~(
    MiniDumpWithoutOptionalData |
    MiniDumpWithFullMemoryInfo |
    MiniDumpWithThreadInfo |
    MiniDumpWithCodeSegs |
    MiniDumpWithPrivateWriteCopyMemory |
    MiniDumpIgnoreInaccessibleMemory |
    MiniDumpWithTokenInformation);

#endif

  HANDLE File = CreateFile(TempName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
  MiniDumpWriteDumpEx(GetCurrentProcess(), GetCurrentProcessId(), File, (MINIDUMP_TYPE) DumpType, &Param, NULL, &Callback);

  CloseHandle(File);

  TerminateProcess(GetCurrentProcess(), 0); // without finalizers
  return EXCEPTION_EXECUTE_HANDLER;
}

static void DebugActivateWER()
{
  MiniDumpWriteDumpEx = (MINIDUMPWRITEDUMP) GetProcAddress(LoadLibrary("dbghelp.DLL"), "MiniDumpWriteDump");
  SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX | SEM_NOOPENFILEERRORBOX);

  if (MiniDumpWriteDumpEx)
    SetUnhandledExceptionFilter(DebugExceptionFilter);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - DEBUGGER - ATTACH
// =======================================================================================================================================================================================================================================================

static void drone_debugger_attach(int force)
{
  if (!IsDebuggerPresent())
  {
    CHAR Params[128];

    STARTUPINFO SI = { }; PROCESS_INFORMATION PI;
    SI.cb = sizeof SI;

    sprintf(Params, "vsjitdebugger.exe -p %lu", GetCurrentProcessId());
    CreateProcess(NULL, Params, NULL, NULL, TRUE, 0, NULL, NULL, &SI, &PI);

    while (!IsDebuggerPresent())
    {
      Sleep(1000);
    }

    DebugBreak();
  }
  else if (force)
  {
    DebugBreak();
  }
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// END OF MODULE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~